"""
CPython 2.7 bytecode opcodes

This is used in bytecode disassembly. This is equivalent to the
opcodes in Python's opcode.py library.
"""

from copy import deepcopy

# These are used from outside this module
from xdis.bytecode import findlinestarts, findlabels

import xdis.opcodes.opcode_2x as opcode_2x
from xdis.opcodes.opcode_2x import def_op, rm_op

# FIXME: can we DRY this even more?

# Make a *copy* of opcode_2x values so we don't pollute 2x

HAVE_ARGUMENT = opcode_2x.HAVE_ARGUMENT
cmp_op = list(opcode_2x.cmp_op)
hasconst = list(opcode_2x.hasconst)
hascompare = list(opcode_2x.hascompare)
hasfree = list(opcode_2x.hasfree)
hasjabs = list(opcode_2x.hasjabs)
hasjrel = list(opcode_2x.hasjrel)
haslocal = list(opcode_2x.haslocal)
hasname = list(opcode_2x.hasname)
hasnargs = list(opcode_2x.hasnargs)
hasvargs = list(opcode_2x.hasvargs)
opmap = deepcopy(opcode_2x.opmap)
opname = deepcopy(opcode_2x.opname)
oppush = list(opcode_2x.oppush)
oppop  = list(opcode_2x.oppop)

EXTENDED_ARG = opcode_2x.EXTENDED_ARG

def updateGlobal(version):
    globals().update({'python_version': version})

    # Canonicalize to PJIx: JUMP_IF_y and POP_JUMP_IF_y
    globals().update({'PJIF': opmap['POP_JUMP_IF_FALSE']})
    globals().update({'PJIT': opmap['POP_JUMP_IF_TRUE']})

    globals().update({'JUMP_OPs': map(lambda op: opname[op], hasjrel + hasjabs)})
    globals().update(dict([(k.replace('+', '_'), v) for (k, v) in opmap.items()]))

def name_op(name, op, pop=-2, push=-2):
    def_op(opname, opmap, name, op, pop, push)
    hasname.append(op)

def jrel_op(name, op, pop=-2, push=-2):
    def_op(opname, opmap, name, op, pop, push)
    hasjrel.append(op)

def jabs_op(name, op):
    def_op(opname, opmap, name, op)
    hasjabs.append(op)

def compare_op(name, op):
    def_op(opname, opmap, name, op)
    hascompare.append(op)

# Bytecodes added since 2.3.
# 2.4
def_op(opname, opmap, 'NOP',           9,  0,  0)
def_op(opname, opmap, 'YIELD_VALUE',  86,  1,  0)

# 2.5
def_op(opname, opmap, 'WITH_CLEANUP', 81, -1, -1)

# 2.6
def_op(opname, opmap, 'STORE_MAP',    54,  3,  2)

# 2.7
rm_op('BUILD_MAP',     104, locals())
rm_op('LOAD_ATTR',     105, locals())
rm_op('COMPARE_OP',    106, locals())
rm_op('IMPORT_NAME',   107, locals())
rm_op('IMPORT_FROM',   108, locals())
rm_op('JUMP_IF_FALSE', 111, locals())
rm_op('EXTENDED_ARG',  143, locals())
rm_op('JUMP_IF_TRUE',  112, locals())

def_op(opname, opmap, 'LIST_APPEND', 94, 2, 1) # Calls list.append(TOS[-i], TOS).
                                               # Used to implement list comprehensions.
def_op(opname, opmap, 'BUILD_SET', 104)        # Number of set items
def_op(opname, opmap, 'BUILD_MAP', 105)
name_op('LOAD_ATTR', 106)
compare_op('COMPARE_OP', 107)

name_op('IMPORT_NAME',         108,  2,  1)  # Index in name list
name_op('IMPORT_FROM',         109,  0,  1)

jabs_op('JUMP_IF_FALSE_OR_POP', 111) # Target byte offset from beginning of code
jabs_op('JUMP_IF_TRUE_OR_POP',  112)  # ""
jabs_op('POP_JUMP_IF_FALSE',    114)  # ""
jabs_op('POP_JUMP_IF_TRUE',     115)  # ""
jrel_op('SETUP_WITH',           143,  0,  2)

def_op(opname, opmap, 'EXTENDED_ARG', 145)
def_op(opname, opmap, 'SET_ADD', 146)
def_op(opname, opmap, 'MAP_ADD', 147)

updateGlobal(2.7)

from xdis import PYTHON_VERSION, IS_PYPY
if PYTHON_VERSION == 2.7 and not IS_PYPY:
    import dis
    # print(set(dis.opmap.items()) - set(opmap.items()))
    # print(set(opmap.items()) - set(dis.opmap.items()))
    assert all(item in opmap.items() for item in dis.opmap.items())
    assert all(item in dis.opmap.items() for item in opmap.items())

# Remove methods so importers aren't tempted to use it.
del name_op, jrel_op, jabs_op



"""
CPython 3.2 bytecode opcodes

This is used in scanner (bytecode disassembly) and parser (Python grammar).

This is a superset of Python 3.2's opcode.py with some opcodes that simplify
parsing and semantic interpretation.
"""

from copy import deepcopy

# These are used from outside this module
from xdis.bytecode import findlinestarts, findlabels

import xdis.opcodes.opcode_3x as opcode_3x
from xdis.opcodes.opcode_3x import fields2copy

# FIXME: can we DRY this even more?

# opmap[opcode_name] => opcode_number
opmap = {}

# opcode[i] => opcode name
opname = [''] * 256

# oppush[op] => number of stack entries pushed
oppush = [0] * 256

# oppop[op] => number of stack entries popped
oppop  = [0] * 256

hasjrel = list(opcode_3x.hasjrel)
hasjabs = []
hasname = list(opcode_3x.hasname)
hasnargs = list(opcode_3x.hasnargs)

for op in range(256): opname[op] = '<%r>' % (op,)
del op

for object in fields2copy:
    globals()[object] =  deepcopy(getattr(opcode_3x, object))

def def_op(opname, opmap, name, op, pop=-2, push=-2):
    opname[op] = name
    opmap[name] = op
    oppush[op] = push
    oppop[op] = pop

def name_op(opname, opmap, name, op, pop=-2, push=-2):
    def_op(opname, opmap, name, op, pop, push)
    hasname.append(op)

def jrel_op(name, op, pop=-2, push=-2):
    def_op(opname, opmap, name, op, pop, push)
    hasjrel.append(op)

# PyPy only
# ----------
name_op(opname, opmap, 'LOOKUP_METHOD', 201, 1, 2)
def_op(opname, opmap, 'CALL_METHOD', 202)
hasnargs.append(202)

# Used only in single-mode compilation list-comprehension generators
def_op(opname, opmap, 'BUILD_LIST_FROM_ARG', 203)

# Used only in assert statements
jrel_op('JUMP_IF_NOT_DEBUG', 204)

# There are no opcodes to remove or change.
# If there were, they'd be listed below.

def updateGlobal():
    globals().update({'python_version': 3.2})

    # JUMP_OPs are used in verification are set in the scanner
    # and used in the parser grammar
    globals().update({'PJIF': opmap['POP_JUMP_IF_FALSE']})
    globals().update({'PJIT': opmap['POP_JUMP_IF_TRUE']})
    globals().update(dict([(k.replace('+', '_'), v) for (k, v) in opmap.items()]))
    globals().update({'JUMP_OPs': map(lambda op: opname[op], hasjrel + hasjabs)})

updateGlobal()

from xdis import PYTHON_VERSION, IS_PYPY
if PYTHON_VERSION == 3.2 and IS_PYPY:
    import dis
    # print(set(dis.opmap.items()) - set(opmap.items()))
    # print(set(opmap.items()) - set(dis.opmap.items()))
    # for item in dis.opmap.items():
    if IS_PYPY:
        assert all(item in opmap.items() for item in dis.opmap.items())
        assert all(item in dis.opmap.items() for item in opmap.items())

# opcode_3x.dump_opcodes(opmap)



"""
CPython 3.0 bytecode opcodes

This is used in scanner (bytecode disassembly) and parser (Python grammar).

This is a superset of Python 3.0's opcode.py with some opcodes that simplify
parsing and semantic interpretation.
"""

from copy import deepcopy

# These are used from outside this module
from xdis.bytecode import findlinestarts, findlabels

import xdis.opcodes.opcode_3x as opcode_3x

from xdis.opcodes.opcode_3x import fields2copy, rm_op

# FIXME: can we DRY this even more?

opmap = {}
opname = [''] * 256

cmp_op     = list(opcode_3x.cmp_op)
hasconst   = list(opcode_3x.hasconst)
hascompare = list(opcode_3x.hascompare)
hasfree    = list(opcode_3x.hasfree)
hasjabs    = list(opcode_3x.hasjabs)
hasjrel    = list(opcode_3x.hasjrel)
haslocal   = list(opcode_3x.haslocal)
hasname    = list(opcode_3x.hasname)
hasnargs   = list(opcode_3x.hasnargs)
hasvargs   = list(opcode_3x.hasvargs)
opmap = deepcopy(opcode_3x.opmap)

oppush = list(opcode_3x.oppush)
oppop  = list(opcode_3x.oppop)

for object in fields2copy:
    globals()[object] =  deepcopy(getattr(opcode_3x, object))

# These are in Python 3.x but not in Python 3.0

rm_op('DUP_TOP_TWO',            5, locals())
rm_op('JUMP_IF_FALSE_OR_POP', 111, locals())
rm_op('JUMP_IF_TRUE_OR_POP',  112, locals())
rm_op('POP_JUMP_IF_FALSE',    114, locals())
rm_op('POP_JUMP_IF_TRUE',     115, locals())
rm_op('DELETE_DEREF',         138, locals())
rm_op('SETUP_WITH',           143, locals())
rm_op('LIST_APPEND',          145, locals())
rm_op('MAP_ADD',              147, locals())

def def_op(name, op, pop=-2, push=-2):
    opname[op] = name
    opmap[name] = op
    oppush[op] = push
    oppop[op] = pop

def jrel_op(name, op, pop=0, push=0):
    def_op(name, op, pop, push)
    hasjrel.append(op)

    # These are are in 3.0 but are not in 3.x or in 3.x they have
# different opcode numbers. Note: As a result of opcode value
# changes, these have to be applied *after* removing ops (with
# the same name).

def_op('ROT_FOUR',        5,  4, 4)
def_op('SET_ADD',        17,  1, 0)
def_op('LIST_APPEND',    18,  2, 1)
def_op('DUP_TOPX',       99)

jrel_op('JUMP_IF_FALSE', 111, 1, 1)
jrel_op('JUMP_IF_TRUE',  112, 1, 1)

# This op is in 3.x but its opcode is a 144 instead
def_op('EXTENDED_ARG',  143)

def updateGlobal():
    # JUMP_OPs are used in verification are set in the scanner
    # and used in the parser grammar
    globals().update({'python_version': 3.0})
    globals().update(dict([(k.replace('+', '_'), v) for (k, v) in opmap.items()]))
    globals().update({'JUMP_OPs': map(lambda op: opname[op], hasjrel + hasjabs)})

updateGlobal()

from xdis import PYTHON_VERSION
if PYTHON_VERSION == 3.0:
    import dis
    # print(set(dis.opmap.items()) - set(opmap.items()))
    # print(set(opmap.items()) - set(dis.opmap.items()))

    assert all(item in dis.opmap.items() for item in opmap.items())
    assert all(item in opmap.items() for item in dis.opmap.items())

# opcode_30.dump_opcodes(opmap)



"""
CPython PYPY 2.7 bytecode opcodes

This is used in bytecode disassembly. This is equivalent to the
opcodes in Python's opcode.py library.
"""

from copy import deepcopy

# These are used from outside this module
from xdis.bytecode import findlinestarts, findlabels

import xdis.opcodes.opcode_27 as opcode_27
from xdis.opcodes.opcode_2x import def_op, rm_op

# FIXME: can we DRY this even more?

# Make a *copy* of opcode_2x values so we don't pollute 2x

HAVE_ARGUMENT = opcode_27.HAVE_ARGUMENT
cmp_op        = list(opcode_27.cmp_op)
hasconst      = list(opcode_27.hasconst)
hascompare    = list(opcode_27.hascompare)
hasfree       = list(opcode_27.hasfree)
hasjabs       = list(opcode_27.hasjabs)
hasjrel       = list(opcode_27.hasjrel)
haslocal      = list(opcode_27.haslocal)
hasname       = list(opcode_27.hasname)
hasnargs      = list(opcode_27.hasnargs)
hasvargs      = list(opcode_27.hasvargs)
opmap         = deepcopy(opcode_27.opmap)
opname        = deepcopy(opcode_27.opname)
EXTENDED_ARG  = opcode_27.EXTENDED_ARG

def updateGlobal(version):
    globals().update({'python_version': version})

    # Canonicalize to PJIx: JUMP_IF_y and POP_JUMP_IF_y
    globals().update({'PJIF': opmap['POP_JUMP_IF_FALSE']})
    globals().update({'PJIT': opmap['POP_JUMP_IF_TRUE']})

    globals().update({'JUMP_OPs': map(lambda op: opname[op], hasjrel + hasjabs)})
    globals().update(dict([(k.replace('+', '_'), v) for (k, v) in opmap.items()]))

def jrel_op(name, op, pop=-2, push=-2):
    def_op(opname, opmap, name, op, pop, push)
    hasjrel.append(op)

def jabs_op(name, op, pop=-2, push=-2):
    def_op(opname, opmap, name, op, pop, push)
    hasjabs.append(op)

def name_op(opname, opmap, name, op, pop=-2, push=-2):
    opcode_27.def_op(opname, opmap, name, op, pop, push)
    hasname.append(op)


# PyPy only
# ----------
name_op(opname, opmap, 'LOOKUP_METHOD', 201, 1, 2)
opcode_27.def_op(opname, opmap, 'CALL_METHOD', 202)
hasnargs.append(202)

# Used only in single-mode compilation list-comprehension generators
def_op(opname, opmap, 'BUILD_LIST_FROM_ARG', 203)

# Used only in assert statements
jrel_op('JUMP_IF_NOT_DEBUG', 204)

# There are no opcodes to remove or change.
# If there were, they'd be listed below.

updateGlobal(2.7)

from xdis import PYTHON_VERSION, IS_PYPY
if PYTHON_VERSION == 2.7 and IS_PYPY:
    import dis
    # print(set(dis.opmap.items()) - set(opmap.items()))
    # print(set(opmap.items()) - set(dis.opmap.items()))
    assert all(item in opmap.items() for item in dis.opmap.items())
    assert all(item in dis.opmap.items() for item in opmap.items())

# Remove methods so importers aren't tempted to use it.
del name_op, jrel_op, jabs_op



"""
CPython PYPY 2.6 bytecode opcodes

This is used in bytecode disassembly. This is equivalent to the
opcodes in Python's opcode.py library.
"""

from copy import deepcopy

# These are used from outside this module
from xdis.bytecode import findlinestarts, findlabels

import xdis.opcodes.opcode_26 as opcode_26

# FIXME: can we DRY this even more?

# Make a *copy* of opcode_2x values so we don't pollute 2x

HAVE_ARGUMENT = opcode_26.HAVE_ARGUMENT
cmp_op        = list(opcode_26.cmp_op)
hasconst      = list(opcode_26.hasconst)
hascompare    = list(opcode_26.hascompare)
hasfree       = list(opcode_26.hasfree)
hasjabs       = list(opcode_26.hasjabs)
hasjrel       = list(opcode_26.hasjrel)
haslocal      = list(opcode_26.haslocal)
hasname       = list(opcode_26.hasname)
hasnargs      = list(opcode_26.hasnargs)
hasvargs      = list(opcode_26.hasvargs)
opmap         = deepcopy(opcode_26.opmap)
opname        = deepcopy(opcode_26.opname)
EXTENDED_ARG  = opcode_26.EXTENDED_ARG

def name_op(opname, opmap, name, op, pop=-2, push=-2):
    opcode_26.def_op(opname, opmap, name, op, pop, push)
    hasname.append(op)


# PyPy only
# ----------
name_op(opname, opmap, 'LOOKUP_METHOD',                201, 1, 2)
opcode_26.def_op(opname, opmap, 'CALL_METHOD',         202)
opcode_26.def_op(opname, opmap, 'BUILD_LIST_FROM_ARG', 203)
opcode_26.def_op(opname, opmap, 'JUMP_IF_NOT_DEBUG',   204)

opcode_26.updateGlobal()

from xdis import PYTHON_VERSION, IS_PYPY
if PYTHON_VERSION == 2.6 and IS_PYPY:
    import dis
    # print(set(dis.opmap.items()) - set(opmap.items()))
    # print(set(opmap.items()) - set(dis.opmap.items()))
    assert all(item in opmap.items() for item in dis.opmap.items())
    assert all(item in dis.opmap.items() for item in opmap.items())
